/*
  Copyright 2018-2021, Barcelona Supercomputing Center (BSC), Spain
  Copyright 2015-2021, Johannes Gutenberg Universitaet Mainz, Germany

  This software was partially supported by the
  EC H2020 funded project NEXTGenIO (Project ID: 671951, www.nextgenio.eu).

  This software was partially supported by the
  ADA-FS project under the SPPEXA project funded by the DFG.

  This file is part of GekkoFS.

  GekkoFS is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  GekkoFS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GekkoFS.  If not, see <https://www.gnu.org/licenses/>.

  SPDX-License-Identifier: GPL-3.0-or-later
*/

#include <regex>
#include <fstream>
#include <random>
#include <utility>

#include <api/info.hpp>
#include <api/connection.hpp>

#include <hermes.hpp>
#include <client/rpc/rpc_types.hpp> // TODO: rpc_types.hpp might need to move
#include <client/env.hpp>           // TODO: independent api/env.hpp

#include <global/env_util.hpp>
#include <global/rpc/distributor.hpp>
#include <global/rpc/rpc_util.hpp>


////////////////////////////////////////////////////////////////////////////////
// helper types and functions                                                 //
////////////////////////////////////////////////////////////////////////////////
namespace {

using namespace gkfs::api;

struct peer_list {

    using endpoint_type = hermes::endpoint;

    peer_list(uint64_t localhost_id, std::vector<endpoint_type> endpoints)
        : localhost_id_(localhost_id), endpoints_(std::move(endpoints)) {}

    uint64_t
    localhost_id() const {
        return localhost_id_;
    }

    hermes::endpoint
    localhost() const {
        return endpoints_.at(localhost_id_);
    }

    std::size_t
    size() const {
        return endpoints_.size();
    }

    uint64_t localhost_id_ = 0;
    std::vector<hermes::endpoint> endpoints_;
};

// TODO: mostly identical to load_hostfile() in preload_util.cpp (could not
//  reuse due to CTX modification in extract_protocol(), which is called from
// load_hostfile()... Consider refactoring.
// std::tuple<rpc::config, std::vector<std::pair<std::string, std::string>>>

std::vector<fs::host>
load_hostfile(const std::string_view& hostfile) {

    //    LOG(DEBUG, "Loading hosts_ file: \"{}\"", hostfile);

    std::ifstream lf(std::string{hostfile});
    if(!lf) {
        throw std::runtime_error(fmt::format("{}", strerror(errno)));
    }
    // std::vector<std::pair<std::string, std::string>> hosts_;
    std::vector<fs::host> hosts;
    const std::regex line_re(R"(^(\S+)\s+(\S+)$)",
                             std::regex::ECMAScript | std::regex::optimize);
    std::string line;
    std::string host;
    std::string uri;
    std::smatch match;
    while(getline(lf, line)) {
        if(!regex_match(line, match, line_re)) {

            //            LOG(ERROR, "Unrecognized line format: [hostfile: '{}',
            //            line: '{}']",
            //                hostfile, line);

            throw std::runtime_error(fmt::format("invalid line: '{}'", line));
        }
        host = match[1];
        uri = match[2];
        hosts.emplace_back(host, uri);
    }
    if(hosts.empty()) {
        throw std::runtime_error(
                "hostfile exists but no addresses could be extracted");
    }

    return hosts;
}

std::vector<fs::host>
find_hosts(const std::string_view& hostfile) {
    try {
        return ::load_hostfile(hostfile);
    } catch(const std::exception& ex) {
        throw std::runtime_error(fmt::format("Error loading hostfile '{}': {}",
                                             hostfile, ex.what()));
    }
}

rpc::config
fetch_rpc_config(const std::string& uri) {

    rpc::config rpc_info;

    if(uri.rfind("://") == std::string::npos) {
        // invalid format. kill client
        throw std::runtime_error(
                fmt::format("Invalid format for URI: '{}'", uri));
    }
    std::string protocol{};

    if(uri.find(gkfs::rpc::protocol::ofi_sockets) != std::string::npos) {
        protocol = gkfs::rpc::protocol::ofi_sockets;
    } else if(uri.find(gkfs::rpc::protocol::ofi_psm2) != std::string::npos) {
        protocol = gkfs::rpc::protocol::ofi_psm2;
    } else if(uri.find(gkfs::rpc::protocol::ofi_verbs) != std::string::npos) {
        protocol = gkfs::rpc::protocol::ofi_verbs;
    }
    // check for shared memory protocol. Can be plain shared memory or real ofi
    // protocol + auto_sm
    if(uri.find(gkfs::rpc::protocol::na_sm) != std::string::npos) {
        if(protocol.empty())
            protocol = gkfs::rpc::protocol::na_sm;
        else
            rpc_info.auto_sm_enabled_ = true;
    }
    if(protocol.empty()) {
        // unsupported protocol. kill client
        throw std::runtime_error(fmt::format(
                "Unsupported RPC protocol found in hosts_ file with URI: '{}'",
                uri));
    }
    //    LOG(INFO,
    //        "RPC protocol '{}' extracted from hosts_ file. Using auto_sm is
    //        '{}'", protocol, CTX->auto_sm());
    rpc_info.protocol_string_ = protocol;

    return rpc_info;
}


/**
 * Looks up a host endpoint via Hermes
 * @param uri
 * @param max_retries
 * @return hermes endpoint, if successful
 * @throws std::runtime_error
 */
// TODO: almost identical to lookup_endpoint() in preload_util.cpp, generalize
hermes::endpoint
lookup_endpoint(const std::unique_ptr<hermes::async_engine>& rpc_service,
                const std::string_view& uri, std::size_t max_retries = 3) {

    //    LOG(DEBUG, "Looking up address \"{}\"", uri);

    std::random_device rd; // obtain a random number from hardware
    std::size_t attempts = 0;
    std::string error_msg;

    do {
        try {
            return rpc_service->lookup(std::string{uri});
        } catch(const std::exception& ex) {
            error_msg = ex.what();

            // LOG(WARNING, "Failed to lookup address '{}'. Attempts [{}/{}] ",
            //    uri, attempts + 1, max_retries);

            // Wait a random amount of time and try again
            std::mt19937 g(rd()); // seed the random generator
            std::uniform_int_distribution<> distr(
                    50, 50 * (attempts + 2)); // define the range
            std::this_thread::sleep_for(std::chrono::milliseconds(distr(g)));
            continue;
        }
    } while(++attempts < max_retries);

    throw std::runtime_error(
            fmt::format("Endpoint for address '{}' could not be found ({})",
                        uri, error_msg));
}


// TODO: mostly identical to init_hermes_client() in preload.cpp
// TODO: consider refactoring them
std::unique_ptr<hermes::async_engine>
create_rpc_service(bool use_auto_sm, const std::string_view& protocol_string) {

    try {

        hermes::engine_options opts{};

        if(use_auto_sm) {
            opts |= hermes::use_auto_sm;
        }

        if(gkfs::rpc::protocol::ofi_psm2 == protocol_string) {
            opts |= hermes::force_no_block_progress;
        }

        return std::make_unique<hermes::async_engine>(
                hermes::get_transport_type(std::string{protocol_string}), opts);
    } catch(const std::exception& ex) {
        throw std::runtime_error(fmt::format(
                "Failed to initialize RPC subsystem: {}\n", ex.what()));
    }
}

std::unique_ptr<hermes::async_engine>
start_rpc_service(const rpc::config& rpc_config) {
    auto rpc_service = create_rpc_service(rpc_config.auto_sm_enabled_,
                                          rpc_config.protocol_string_);
    rpc_service->run();
    return rpc_service;
}

// TODO: mostly identical to connec_to_hosts() in rpc_util.cpp
peer_list
resolve(const std::unique_ptr<hermes::async_engine>& rpc_service,
        const std::vector<fs::host>& hosts) {

    const auto localhost_name = gkfs::rpc::get_my_hostname(true);
    uint64_t localhost_id = 0;
    bool localhost_found = false;

    std::vector<hermes::endpoint> endpoints;
    endpoints.resize(hosts.size());

    std::vector<uint64_t> host_ids(hosts.size());
    // populate vector with [0, ..., host_size - 1]
    std::iota(std::begin(host_ids), std::end(host_ids), 0);
    /*
     * Shuffle hosts_ to balance addr lookups to all hosts_
     * Too many concurrent lookups send to same host
     * could overwhelm the server,
     * returning error when addr lookup
     */
    std::random_device rd; // obtain a random number from hardware
    std::mt19937 g(rd());  // seed the random generator
    std::shuffle(host_ids.begin(), host_ids.end(), g); // Shuffle hosts_
    // vector
    // lookup addresses and put abstract server addresses into rpc_addresses

    for(const auto& id : host_ids) {
        const auto& hostname = hosts.at(id).hostname();
        const auto& uri = hosts.at(id).uri();

        endpoints[id] = ::lookup_endpoint(rpc_service, uri);

        if(!localhost_found && hostname == localhost_name) {
            //                LOG(DEBUG, "Found local host: {}", hostname);
            localhost_id = id;
            localhost_found = true;
        }

        //            LOG(DEBUG, "Found peer: {}",
        //            endpoints[id].to_string());
    }

    return peer_list{localhost_id, endpoints};
}

// TODO: mostly identical to gkfs::rpc::forward_get_fs_config() in
// TODO: forward_management.cpp, consider refactoring them
fs::config
fetch_fs_config(const std::unique_ptr<hermes::async_engine>& rpc_service,
                const peer_list& peers) {

    const auto endp = peers.localhost();
    gkfs::rpc::fs_config::output out;

    try {
        //        LOG(DEBUG, "Retrieving file system configurations from
        //        daemon");
        // TODO(amiranda): add a post() with RPC_TIMEOUT to hermes so that
        // we can retry for RPC_TRIES (see old commits with margo)
        // TODO(amiranda): hermes will eventually provide a post(endpoint)
        // returning one result and a broadcast(endpoint_set) returning a
        // result_set. When that happens we can remove the .at(0) :/
        out = rpc_service->post<gkfs::rpc::fs_config>(endp).get().at(0);
    } catch(const std::exception& ex) {
        throw std::runtime_error(
                "Error retrieving gekkofs_instance configurations from "
                "daemon");
    }

    return {out.atime_state(),  out.mtime_state(),
            out.ctime_state(),  out.link_cnt_state(),
            out.blocks_state(), out.uid(),
            out.gid(),          out.rootdir(),
            out.mountdir()};
}


} // namespace

namespace gkfs::api::fs {

class connection::impl {

public:
    impl(const std::string_view& hostfile)
        : hosts_(find_hosts(hostfile)),
          rpc_config_(fetch_rpc_config(hosts_.at(0).uri())),
          rpc_service_(start_rpc_service(rpc_config_)),
          peers_(resolve(rpc_service_, hosts_)),
          fs_config_(fetch_fs_config(rpc_service_, peers_)),
          distributor_(std::make_unique<gkfs::rpc::SimpleHashDistributor>(
                  peers_.localhost_id(), peers_.size())) {

        using namespace std::string_literals;

#if 0
        // read host information
        std::vector<std::pair<std::string, std::string>> hosts_{};

        try {
            //        LOG(INFO, "Loading peer addresses...");
            std::tie(rpc_config_, hosts_) = ::find_hosts();
        } catch(const std::exception& e) {
            throw std::runtime_error(
                    fmt::format("Failed to load hosts_ addresses:"
                                " {}",
                                e.what()));
        }
#endif

        // rpc_service_ = create_rpc_service();
        // rpc_service_->run();

#if 0
        try {
            ping(hosts_);
        } catch(const std::exception& e) {
            throw std::runtime_error(
                    fmt::format("Failed to connect to hosts_: {}\n", e.what()));
        }
#endif

#if 0
        distributor_ = std::make_unique<gkfs::rpc::SimpleHashDistributor>(
                local_host_id_, host_endpoints_.size());
#endif

        // LOG(INFO, "Retrieving file system configuration...");
#if 0
        try {
            fs_config_ = fetch_fs_info();
        } catch(const std::exception& ex) {
            throw std::runtime_error(
                    fmt::format("Unable to fetch file system "
                                "configurations from daemon process "
                                "through RPC: {}",
                                ex.what()));
        }
#endif

        // LOG(INFO, "Environment initialization successful.");
    }

    impl(const impl& other) = delete;
    impl(impl&& rhs) = default;
    impl&
    operator=(const impl& other) = delete;
    impl&
    operator=(impl&& rhs) = default;
    ~impl() = default;

private:
    ////////////////////////////////////////////////////////////////////////////
    // members                                                                //
    ////////////////////////////////////////////////////////////////////////////
    std::vector<fs::host> hosts_;
    rpc::config rpc_config_;
    std::unique_ptr<hermes::async_engine> rpc_service_;
    peer_list peers_;
    std::vector<hermes::endpoint> host_endpoints_;
    fs::config fs_config_;
    std::unique_ptr<gkfs::rpc::Distributor> distributor_;
};

connection::connection()
    : pimpl_(new connection::impl(gkfs::env::get_var(
              gkfs::env::HOSTS_FILE, gkfs::config::hostfile_path))) {}

connection::connection(const std::string& hostfile)
    : pimpl_(new connection::impl(hostfile)) {}

connection::~connection() = default;

} // namespace gkfs::api::fs
