/*
  Copyright 2018-2021, Barcelona Supercomputing Center (BSC), Spain
  Copyright 2015-2021, Johannes Gutenberg Universitaet Mainz, Germany

  This software was partially supported by the
  EC H2020 funded project NEXTGenIO (Project ID: 671951, www.nextgenio.eu).

  This software was partially supported by the
  ADA-FS project under the SPPEXA project funded by the DFG.

  This file is part of GekkoFS.

  GekkoFS is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  GekkoFS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GekkoFS.  If not, see <https://www.gnu.org/licenses/>.

  SPDX-License-Identifier: GPL-3.0-or-later
*/

#include <filesystem>
#include <fstream>

#include <catch2/catch.hpp>
#include <fmt/format.h>

#include <api/gekkofs.hpp>

// A temporary file with RAII removal
struct temporary_file {
    temporary_file(std::filesystem::path filename)
        : filename_(std::move(filename)), ofs_(filename_) {}

    ~temporary_file() {
        ofs_.close();
        std::filesystem::remove(filename_);
    }

    void
    write(const std::string& text) {
        ofs_ << text;
        ofs_.flush();
    }

    std::filesystem::path
    filename() const {
        return filename_;
    }

    std::filesystem::path filename_;
    std::ofstream ofs_;
};

SCENARIO(" ", "[api]") {

    using namespace std::string_literals;
    using Catch::Matchers::Contains;
    using namespace gkfs::api;

    GIVEN(" a fs::connection ") {

        WHEN(" hostfile does not exist ") {

            THEN(" an exception is thrown ") {
                REQUIRE_THROWS_WITH(fs::connection(),
                                    Contains("No such file or directory"s));
            }
        }

        WHEN(" hostfile is empty ") {

            temporary_file tmp{"empty.hostfile"};

            THEN(" an exception is thrown ") {
                REQUIRE_THROWS_WITH(
                        fs::connection(tmp.filename()),
                        Contains("hostfile exists but no addresses could be "
                                 "extracted"s));
            }
        }

        WHEN(" hostfile is invalid ") {

            temporary_file tmp{"invalid.hostfile"};

            tmp.write("foobarbaz\nbarfoobaz\nbazbarfoo\n");

            THEN(" an exception is thrown ") {
                REQUIRE_THROWS_WITH(fs::connection(tmp.filename()),
                                    Contains("invalid line: 'foobarbaz'"s));
            }
        }

        WHEN(" hostfile is partially invalid ") {

            temporary_file tmp{"pinvalid.hostfile"};

            tmp.write(
                    "host.example.com ofi+sockets://127.0.0.1:11586\nfoobarbaz\nbarfoobaz\nbazbarfoo\n");

            THEN(" an exception is thrown ") {
                REQUIRE_THROWS_WITH(fs::connection(tmp.filename()),
                                    Contains("invalid line: 'foobarbaz'"s));
            }
        }

        WHEN(" hostfile is valid ") {

            temporary_file tmp{"valid.hostfile"};

            // tmp.write("host.example.com ofi+sockets://127.0.0.1:11586\n");
            THEN(" a connection instance is constructed ") {
                tmp.write("athenai ofi+sockets://127.0.0.1:11586\n");

                REQUIRE_NOTHROW(fs::connection(tmp.filename()));
            }
        }
    }
}
